//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#ifdef _WIN32
	#include <windows.h>
#else
	#include <stdarg.h>
#endif

#include "AnsiString.h"
#include <list>

#include <audiere.h>

//using namespace audiere;
//using namespace std::list;


#define CUSTOM_CONTROL_ID       1011


INVOKER RemoteInvoker=0;

typedef struct {
	AnsiString key;
	audiere::OutputStreamPtr stream;
} MyStream;

std::list<MyStream *> lst_stream;

audiere::OutputStreamPtr GetStream(AnsiString key) {
	audiere::OutputStreamPtr stream=0;

	std::list<MyStream *>::const_iterator it;
	for(it=lst_stream.begin(); it!=lst_stream.end(); ++it) {
		if ((*it)->key==key) {
			stream=(*it)->stream;
			break;
		}
	}
	return stream;
}

void AddStream(AnsiString key, audiere::OutputStreamPtr stream) {
	MyStream *ns=new MyStream;
	ns->key=key;
	ns->stream=stream;

	lst_stream.push_front(ns);
}

audiere::OutputStreamPtr CreateStream(AnsiString key, char *filename) {
    audiere::AudioDevicePtr device(audiere::OpenDevice());
	audiere::OutputStreamPtr stream(audiere::OpenSound(device, filename, false));
	bool found_flag=false;
	std::list<MyStream *>::const_iterator it;
	for(it=lst_stream.begin(); it!=lst_stream.end(); ++it) {
		if ((*it)->key==key) {
			//(*it)=stream
			lst_stream.remove(*it);
			found_flag=true;
			break;
		}
	}
	if (!found_flag) {
		AddStream(key, stream);
	}
	return stream;
}


audiere::OutputStreamPtr RemStream(AnsiString key) {
	audiere::OutputStreamPtr stream=0;

	std::list<MyStream *>::const_iterator it;
	for(it=lst_stream.begin(); it!=lst_stream.end(); ++it) {
		if ((*it)->key==key) {
			stream=(*it)->stream;
			lst_stream.remove((*it));
			break;
		}
	}
	return stream;
}

void Clean() {
	std::list<MyStream *>::const_iterator it;
	for(it=lst_stream.begin(); it!=lst_stream.end(); ++it) {
		delete (*it);
	}
	lst_stream.clear();
}

PREFIX int Invoke(void *context,int INVOKE_TYPE,...) {
    va_list arg;
    switch (INVOKE_TYPE) {
        case INVOKE_QUERYTEST:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id=va_arg(arg, int);
                va_end(arg);

                if (cls_id==CUSTOM_CONTROL_ID)
                    return E_ACCEPTED;
            }
            break;
		case INVOKE_STATICQUERY:
			{
                va_start(arg, INVOKE_TYPE);
                int cls_id=va_arg(arg, int);
				char *param1=va_arg(arg, char *);
				int param1_len=va_arg(arg, int);
				char *param2=va_arg(arg, char *);
				int param2_len=va_arg(arg, int);
                va_end(arg);

				if (cls_id!=CUSTOM_CONTROL_ID)
                    break;
				AnsiString key(param1);
				AnsiString command;
				AnsiString parameter;
				bool separator_found=false;

				//std::cerr << "Line length: " << param2_len << "\n";
				for (int i=0;i<param2_len;i++) {
					char c=param2[i];
					if (separator_found) {
						parameter+=c;
					} else {
						if (c==':')
							separator_found=true;
						else
							command+=c;
					}
				}
				//std::cerr << "LINE: " << param2 << "\n";
				//std::cerr << "CMD: " << command.c_str() << "\n";
				//std::cerr << "Param: " << parameter.c_str() << "\n";
				
				if (command==(char *)"volume") {
					audiere::OutputStreamPtr stream=GetStream(key);
					if (stream) {
						stream->setVolume(parameter.ToFloat());
					}
				} else
				if (command==(char *)"repeat") {
					audiere::OutputStreamPtr stream=GetStream(key);
					if (stream) {
						stream->setRepeat(parameter.ToInt());
					}
				} else
				if (command==(char *)"play") {
					audiere::OutputStreamPtr stream=GetStream(key);
					if (stream) {
						stream->play();
					}
				} else
				if (command==(char *)"reset") {
					audiere::OutputStreamPtr stream=GetStream(key);
					if (stream) {
						stream->reset();
					}
				} else
				if (command==(char *)"stop") {
					audiere::OutputStreamPtr stream=GetStream(key);
					if (stream) {
						stream->stop();
					}
				} else
				if (command==(char *)"done") {
					audiere::OutputStreamPtr stream=RemStream(key);
					if (stream) {
						stream->stop();
						// it's a smart pointer ! NO DELETE !
						stream=0;
					}
				} else
				if (command==(char *)"open") {
					audiere::OutputStreamPtr stream=CreateStream(key, parameter);
				} else
				if (command==(char *)"pitch") {
					audiere::OutputStreamPtr stream=GetStream(key);
					if (stream) {
						stream->setPitchShift(parameter.ToFloat());
					}
				} else
				if (command==(char *)"pan") {
					audiere::OutputStreamPtr stream=GetStream(key);
					if (stream) {
						stream->setPan(parameter.ToFloat());
					}
				} else
				if (command==(char *)"position") {
					audiere::OutputStreamPtr stream=GetStream(key);
					if (stream) {
						stream->setPosition(parameter.ToInt());
					}
				}
                return E_ACCEPTED;
			}
            break;
		case INVOKE_STATICRESPONSE:
			{
                va_start(arg, INVOKE_TYPE);
                int cls_id=va_arg(arg, int);
				char *param1=va_arg(arg, char *);
				int param1_len=va_arg(arg, int);
				char *param2=va_arg(arg, char *);
				int param2_len=va_arg(arg, int);

				char **result=va_arg(arg, char **);
				int *result_len=va_arg(arg, int *);

				va_end(arg);

				if (cls_id!=CUSTOM_CONTROL_ID)
                    break;

                return E_ACCEPTED;
			}
            break;
        case INVOKE_SETEVENT:
            break;
        case INVOKE_SETPROPERTY:
            break;
        case INVOKE_GETPROPERTY:
            break;
        case INVOKE_SETSTREAM:
            break;
        case INVOKE_SETSECONDARYSTREAM:
            // not needed
            break;
        case INVOKE_CUSTOMMESSAGE:
            break;
        case INVOKE_CREATE:
            break;
        case INVOKE_DESTROY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void *handler  =   va_arg(arg, void *);
                va_end(arg);

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;
                
                return E_ACCEPTED;
            }
            break;
        case INVOKE_INITLIBRARY:
            va_start(arg, INVOKE_TYPE);
            RemoteInvoker     =   va_arg(arg, INVOKER);
            va_end(arg);
            return E_NOERROR;

        case INVOKE_DONELIBRARY:
            return E_NOERROR;

        default:
            return E_NOTIMPLEMENTED;
    }
    return E_NOTIMPLEMENTED;
}
