//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include <stdarg.h>
#ifdef _WIN32
	#include <windows.h>
    #include <fcntl.h>
    //#define pipe(phandles)	_pipe (phandles, 4096, _O_BINARY)
#endif

#include "AnsiString.h"
#include "semhh.h"

#include <portaudio.h>
#include <unistd.h>
#include <stdio.h>
#include <speex/speex.h> 

#define CUSTOM_CONTROL_ID       1016

#define P_SAMPLERATE            1000
#define P_NUMCHANNELS           1001
#define P_RECORD                1002
#define P_ADDBUFFER             1003
#define P_BEGINPLAYBACK         1004
#define P_QUALITY               1005
#define P_USECOMPRESSION        1006

// 2 seconds
#define MAX_LATENCY_BUF         2 * (SAMPLE_RATE*NUM_CHANNELS)
#define MAX_BUFFER              0x3FFFF

INVOKER RemoteInvoker=0;
AnsiString buffered_text;

#define SAMPLE short
#define PA_SAMPLE_TYPE  paInt16
#define SAMPLE_SILENCE  (0)
#define FRAMES_PER_BUFFER (1024)

#define EVENT_ON_BUFFER         350

static int recordCallback( const void *inputBuffer, void *outputBuffer,
                           unsigned long framesPerBuffer,
                           const PaStreamCallbackTimeInfo* timeInfo,
                           PaStreamCallbackFlags statusFlags,
                           void *userData );

static int playCallback( const void *inputBuffer, void *outputBuffer,
                         unsigned long framesPerBuffer,
                         const PaStreamCallbackTimeInfo* timeInfo,
                         PaStreamCallbackFlags statusFlags,
                         void *userData );

//static THREAD_TYPE EncodingLoop(LPVOID ao);

class AudioObject {
public:
    PaStreamParameters  inputParameters;
    PaStreamParameters  outputParameters;
    PaStream*           stream;
    PaError             err;
    int                 SAMPLE_RATE;
    int                 NUM_CHANNELS;
    int                 finished;
    int                 quality;
    int                 speex_frame_size;
    bool                on_buffer_full;
    SpeexBits           bits;
    void                *enc_state;
    void                *dec_state; 
    AnsiString          audio_buf;
    SAMPLE              propagation[0xFFF];
    int                 propagation_size;
    bool                use_compression;
    HHSEM               lock;

    AudioObject() {
        err = paNoError;
        stream=0;
        SAMPLE_RATE=44100;
        NUM_CHANNELS=1;
        finished=0;
        on_buffer_full=false;

        enc_state=0;
        dec_state=0;
        quality=8;
        speex_frame_size=0;
        seminit(lock, 1);
        propagation_size=0;
        use_compression=true;
    }

    ~AudioObject() {
    }

    bool IsActive() {
        if (!stream)
            return false;
        if ( Pa_IsStreamActive( stream ) == 1 )
            return true;
        return false;
    }

    bool Stop() {
        finished=0;
        if (!stream)
            return true;

        err = Pa_CloseStream( stream );

        if (dec_state) {
            speex_decoder_destroy(dec_state); 
            dec_state=0;
        }
        if (enc_state) {
            speex_encoder_destroy(enc_state); 
            enc_state=0;
        }
        speex_bits_destroy(&bits);
        stream=0;
        return true;
    }

    void SpeexDeBuffer(char *buf, int buf_size) {
        if (!stream)
            return;
        spx_int16_t internal2[0xFFFF];
        if ((!buf) || (!buf_size))
            return;

        int n_frames=(unsigned char)buf[0];
        if (!n_frames)
            return;
        buf++;
        buf_size--;
        int p_size=(unsigned char)buf[0];//buf_size/n_frames;
        buf++;
        buf_size--;
        int total=0;
        fprintf(stderr, "Speex: init packet: %i frames of %i size\n",n_frames,p_size);
        fflush(stderr);
        do {
            n_frames--;
            if (n_frames<0) {
                n_frames=(unsigned char)buf[0];
                buf++;
                buf_size--;
                p_size=(unsigned char)buf[0];
                buf++;
                buf_size--;
                n_frames--;
            }
            speex_bits_reset(&bits);
            speex_bits_read_from(&bits, buf, buf_size); 
            int err=speex_decode_int(dec_state, &bits, (spx_int16_t *)&internal2[total]); 
            if (err) {
                fprintf(stderr, "Speex: Error decoding: %i\n",err);
                fflush(stderr);
            }
            total+=speex_frame_size;
            buf+=p_size;
            buf_size-=p_size;
            if (buf_size<p_size)
                break;
        } while ((!err) /*&& (buf_size>=p_size)*/);
        semp(lock);
        AnsiString temp;
        if (audio_buf.Length()>MAX_LATENCY_BUF) {
            audio_buf.LoadBuffer((char *)internal2, total*sizeof(short));
        } else {
            temp.LoadBuffer((char *)internal2, total*sizeof(short));
            audio_buf+=temp;
        }
        semv(lock);
    }

    int SpeexBuffer2(const SAMPLE *buf, int buf_size) {
        if (!stream)
            return -1;
        
        // buffer to small, will ignore it
        // that's not good
        // REWRITE THIS

        // at least 3 packages !
        if (buf_size+propagation_size<speex_frame_size*3) {
            propagation_size=0;
            return -2;
        }

        char internal2[0xFFFF];

        int total=0;
        int pos=0;
        
        //speex_bits_reset(&bits);
        //speex_encode_int(enc_state, (spx_int16_t *)&buf[pos], &bits);

        internal2[0]=0;//packets !
        total++;
        internal2[0]=0;//p_size
        total++;
        if (propagation_size) {
            int delta=speex_frame_size-propagation_size;
            for (int i=0;i<delta;i++)
                propagation[propagation_size+i]=buf[i];
            pos=delta;
            buf_size-=delta;
            speex_bits_reset(&bits);
            speex_encode_int(enc_state, (spx_int16_t *)propagation, &bits);
            int b=speex_bits_write(&bits, &internal2[total], 0xFFFF-total);
            if (b>0) {
                total+=b;
                internal2[1]=b;
                propagation_size=0;
                internal2[0]++;
            }
        }

        while (buf_size>=speex_frame_size) {
            speex_bits_reset(&bits);
            speex_encode_int(enc_state, (spx_int16_t *)&buf[pos], &bits);
            int nbBytes = speex_bits_write(&bits, &internal2[total], 0xFFFF-total);
            if (nbBytes<=0)
                break;
            internal2[1]=nbBytes;
            pos+=speex_frame_size;
            buf_size-=speex_frame_size;
            total+=nbBytes;
            internal2[0]++;
        }
        if (buf_size) {
            for (int i=0;i<buf_size;i++)
                propagation[i]=buf[pos+i];
            propagation_size=buf_size;
        }
        RemoteInvoker(this, INVOKE_FIREEVENT2, (int)EVENT_ON_BUFFER, internal2, total);
        return buf_size;
    }

    bool Record(int device=-1) {
        if ((stream) && (IsActive()))
            Stop();
        if ((NUM_CHANNELS==1) && (use_compression)) {
            speex_bits_init(&bits);
            enc_state = speex_encoder_init(&speex_nb_mode);
            speex_encoder_ctl(enc_state,SPEEX_SET_QUALITY,&quality);
            speex_encoder_ctl(enc_state,SPEEX_SET_SAMPLING_RATE, &SAMPLE_RATE);
            speex_encoder_ctl(enc_state,SPEEX_GET_FRAME_SIZE,&speex_frame_size);
        }
        inputParameters.device = device<0 ? Pa_GetDefaultInputDevice() : device;
        if (inputParameters.device == paNoDevice)
            return false;
        inputParameters.channelCount = NUM_CHANNELS;
        inputParameters.sampleFormat = PA_SAMPLE_TYPE;
        inputParameters.suggestedLatency = Pa_GetDeviceInfo( inputParameters.device )->defaultLowInputLatency;
        inputParameters.hostApiSpecificStreamInfo = NULL;

        err = Pa_OpenStream(
                  &stream,
                  &inputParameters,
                  NULL,                  /* &outputParameters, */
                  SAMPLE_RATE,
                  FRAMES_PER_BUFFER,
                  paClipOff,      /* we won't output out of range samples so don't bother clipping them */
                  recordCallback,
                  this );

        err = Pa_StartStream( stream );
        if ( err != paNoError )
            return false;
        return true;
    }

    bool Play(int device=-1) {
        if ((stream) && (IsActive()))
            Stop();

        if ((NUM_CHANNELS==1) && (use_compression)) {
            speex_bits_init(&bits);
            dec_state = speex_decoder_init(&speex_nb_mode); 
            speex_decoder_ctl(dec_state,SPEEX_SET_SAMPLING_RATE, &SAMPLE_RATE); 
            speex_decoder_ctl(dec_state,SPEEX_GET_FRAME_SIZE,&speex_frame_size); 
        }

        outputParameters.device = device<0 ? Pa_GetDefaultOutputDevice() : device;
        if (inputParameters.device == paNoDevice)
            return false;
        outputParameters.channelCount = NUM_CHANNELS;
        outputParameters.sampleFormat =  PA_SAMPLE_TYPE;
        outputParameters.suggestedLatency = Pa_GetDeviceInfo( outputParameters.device )->defaultLowOutputLatency;
        outputParameters.hostApiSpecificStreamInfo = NULL;

        err = Pa_OpenStream(
                  &stream,
                  NULL,
                  &outputParameters,
                  SAMPLE_RATE,
                  FRAMES_PER_BUFFER,
                  paClipOff,
                  playCallback,
                  this );

        err = Pa_StartStream( stream );
        if ( err != paNoError )
            return false;
        return true;
    }

    void AddBuffer(void *str_buffer, int buf_len) {
        if (!stream)
            return;
        if ((NUM_CHANNELS==1) && (use_compression)) {
            SpeexDeBuffer((char *)str_buffer, buf_len);
        } else {
            semp(lock);
            if (audio_buf.Length()>MAX_LATENCY_BUF) {
                audio_buf.LoadBuffer((char *)str_buffer, buf_len);
            } else {
                AnsiString temp;
                temp.LoadBuffer((char *)str_buffer, buf_len);
                audio_buf+=temp;
            }
            semv(lock);
        }
    }
};

static int playCallback( const void *inputBuffer, void *outputBuffer,
                         unsigned long framesPerBuffer,
                         const PaStreamCallbackTimeInfo* timeInfo,
                         PaStreamCallbackFlags statusFlags,
                         void *userData )
{
    AudioObject *data = (AudioObject*)userData;
    SAMPLE *wptr = (SAMPLE*)outputBuffer;
    unsigned long size=framesPerBuffer*data->NUM_CHANNELS*sizeof(SAMPLE);

    int finished=data->finished;
    int rsize=0;
    int rd=0;
    int left=size;
    //while (left) {
    if (left) {
        semp(data->lock);
        int len=data->audio_buf.Length();
        if (len>=left) {
            memcpy(wptr, data->audio_buf.c_str(), left);
            if (len==left)
                data->audio_buf=(char *)"";
            else {
                AnsiString temp;
                temp.LoadBuffer(data->audio_buf.c_str()+left, len-left);
                data->audio_buf=temp;
            }
            left=0;
        }
        semv(data->lock);
    }
    return finished;
}

static int recordCallback( const void *inputBuffer, void *outputBuffer,
                           unsigned long framesPerBuffer,
                           const PaStreamCallbackTimeInfo* timeInfo,
                           PaStreamCallbackFlags statusFlags,
                           void *userData )
{
    AudioObject *data = (AudioObject*)userData;
    const SAMPLE *rptr = (const SAMPLE*)inputBuffer;
    int size2=framesPerBuffer*data->NUM_CHANNELS;
    int size=size2*sizeof(SAMPLE);
    int finished=data->finished;

    if (data->on_buffer_full) {
        if ((data->NUM_CHANNELS==1) && (data->use_compression)) {
            data->SpeexBuffer2(rptr, size2);
        } else
            RemoteInvoker(data, INVOKE_FIREEVENT2, (int)EVENT_ON_BUFFER, rptr, size);
    }
    return finished;
}


PREFIX int Invoke(void *context,int INVOKE_TYPE,...) {
    va_list arg;
    switch (INVOKE_TYPE) {
        case INVOKE_QUERYTEST:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id=va_arg(arg, int);
                va_end(arg);

                if (cls_id==CUSTOM_CONTROL_ID)
                    return E_ACCEPTED;
            }
            break;
        case INVOKE_SETEVENT:
            {
                if (!RemoteInvoker)
                    break;
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int event_id    =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                AudioObject *ctx=(AudioObject *)context;
                switch (event_id) {
                    case EVENT_ON_BUFFER:
                        ctx->on_buffer_full=true;
                        return E_ACCEPTED;
                        break;
                }
            }
            break;
        case INVOKE_SETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);
                AnsiString param(str_buffer);
                int param_int=param.ToInt();

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                AudioObject *ctx=(AudioObject *)context;
                switch (prop_id) {
                    case P_SAMPLERATE:
                        ctx->SAMPLE_RATE=param.ToInt();
                        break;
                    case P_NUMCHANNELS:
                        ctx->NUM_CHANNELS=param.ToInt();
                        break;
                    case P_RECORD:
                        if (param_int==-2) {
                            ctx->Stop();
                        } else
                            ctx->Record(param_int);
                        break;
                    case P_ADDBUFFER:
                        ctx->AddBuffer(str_buffer, buf_len);
                        break;
                    case P_BEGINPLAYBACK:
                        if (param_int==-2) {
                            ctx->Stop();
                        } else
                            ctx->Play(param_int);
                        break;
                    case P_QUALITY:
                        ctx->quality=param.ToInt();
                        break;
                    case P_USECOMPRESSION:
                        ctx->use_compression=param.ToInt();
                        break;

                }
            }
            break;
       case INVOKE_GETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char **str_buffer=   va_arg(arg, char **);
                int *buf_len     =   va_arg(arg, int*);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                buffered_text=(char *)"";
                *str_buffer=buffered_text.c_str();
                *buf_len=buffered_text.Length();
                return E_NOERROR;
            }
            break;
        case INVOKE_SETSTREAM:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                char *target    =   va_arg(arg, char *);
                int len_target  =   va_arg(arg, int);
                char *value     =   va_arg(arg, char *);
                int len_value   =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

            }
            break;
        case INVOKE_SETSECONDARYSTREAM:
            // not needed
            break;
        case INVOKE_CUSTOMMESSAGE:
            {
            }
            break;
        case INVOKE_CREATE:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void **handler  =   va_arg(arg, void **);
                int *pack_type  =   va_arg(arg, int *);
                va_end(arg);
                *pack_type=(int)2;

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;

                AudioObject *ctx=new AudioObject();
                *handler=ctx;
                return E_ACCEPTED_NOADD;
            }
            break;
        case INVOKE_DESTROY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void *handler  =   va_arg(arg, void *);
                va_end(arg);

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;

                return E_ACCEPTED;
            }
            break;
        case INVOKE_INITLIBRARY:
            va_start(arg, INVOKE_TYPE);
            RemoteInvoker     =   va_arg(arg, INVOKER);
            va_end(arg);
            Pa_Initialize();
            return E_NOERROR;
        case INVOKE_DONELIBRARY:
            return E_NOERROR;
        default:
            return E_NOTIMPLEMENTED;
    }
    return E_NOTIMPLEMENTED;
}


