//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include <stdarg.h>
#ifdef _WIN32
	#include <windows.h>
#endif

#include <gtk/gtk.h>
#include <poppler.h>
#include "AnsiString.h"

#define CUSTOM_CONTROL_ID       1010

#define P_PDFPAGES              10001
#define P_PDFCURRENTPAGE        10002
#define P_PDFZOOM               10003
#define P_PDFCURRENTPAGELABEL   10004
#define P_PAGEICONVIEW          10005
#define P_THUMBSIZE             10006
#define P_MARKTEXT              10007


INVOKER RemoteInvoker=0;
class MyPDFData {
private:
    GtkListStore *m_refListModel;

    void prepare() {
        if (!iconview)
            return;
        if (prepared)
            return;

        prepared=true;
        gtk_icon_view_set_model(iconview, GTK_TREE_MODEL(m_refListModel));
        gtk_icon_view_set_pixbuf_column(iconview, 1);
        gtk_icon_view_set_markup_column(iconview, 0);
    }

    void Add(GdkPixbuf *pixbufmm, char *description) {
        GtkTreeIter iter;
        gtk_list_store_append(m_refListModel, &iter);
        gtk_list_store_set(m_refListModel, &iter, (int)1, pixbufmm, (int)0, description, -1);
    }
public:
    PopplerDocument *pdoc;
    PopplerPage *ppage;
    double zoom;
    int thumb_size;
    AnsiString master_data;
    GtkIconView *iconview;
    bool prepared;

    void Load(char *data, int data_len) {
        master_data.LoadBuffer(data, data_len);
    }

    void PopulateIconView() {
        if ((!iconview) || (!pdoc))
            return;

        gtk_list_store_clear(m_refListModel);
        prepare();

        int pages=poppler_document_get_n_pages(pdoc);
        for (int i=0;i<pages;i++) {
            PopplerPage  *page=poppler_document_get_page(pdoc,i);
            if (page) {
                double src_width=0;
                double src_height=0;
                poppler_page_get_size(page,&src_width,&src_height);

                if (src_width==0)
                    src_width=1;
                if (src_height==0)
                    src_height=1;

                double thumb_zoom=1;
                if (src_width>src_height) {
                    thumb_zoom=thumb_size/src_width;
                } else {
                    thumb_zoom=thumb_size/src_height;
                }

                GdkPixbuf *pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, false, 8, (int)src_width*thumb_zoom, (int)src_height*thumb_zoom);
                poppler_page_render_to_pixbuf(page,0,0,(int)src_width*thumb_zoom,(int)src_height*thumb_zoom,thumb_zoom,0,pixbuf); 

                AnsiString S;
                S+=(char *)"<b>";
                S+=AnsiString((long)(i+1));
                S+=(char *)"</b>";
                Add(pixbuf, S.c_str());
            }
        }
    }

    MyPDFData() {
        zoom=1.0;
        pdoc=0;
        ppage=0;
        m_refListModel  = gtk_list_store_new(2, G_TYPE_STRING, GDK_TYPE_PIXBUF);
        //iconview=GTK_ICON_VIEW(gtk_icon_view_new_with_model(GTK_TREE_MODEL(m_refListModel)));
        prepared=false;
        thumb_size=150;
    }
};

// Control invoker function
AnsiString buffered_text;
GdkPixbuf *pixbuf_main;

PREFIX int Invoke(void *context,int INVOKE_TYPE,...) {
    va_list arg;
    switch (INVOKE_TYPE) {
        case INVOKE_QUERYTEST:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id=va_arg(arg, int);
                va_end(arg);

                if (cls_id==CUSTOM_CONTROL_ID)
                    return E_ACCEPTED;
            }
            break;
        case INVOKE_SETEVENT:
            {
                if (!RemoteInvoker)
                    break;
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int event_id    =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

            }
            break;
        case INVOKE_SETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);
                AnsiString param(str_buffer);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                MyPDFData *pdata=0;
                RemoteInvoker(context,INVOKE_GETEXTRAPTR,&pdata);

                if (!pdata) {
                    pdata=new MyPDFData();
                    RemoteInvoker(context,INVOKE_SETEXTRAPTR,(void*)pdata);
                }

                switch (prop_id) {
                        //case P_PDFPAGES:
                        case P_PDFCURRENTPAGE:
                             if (pdata->pdoc) {
                                if (pdata->ppage)
                                    g_object_unref(pdata->ppage);

                                PopplerPage *ppage=poppler_document_get_page(pdata->pdoc,param.ToInt());
                                pdata->ppage=ppage;
                                if (ppage) {
                                    double src_width=0;
                                    double src_height=0;
                                    poppler_page_get_size(pdata->ppage,&src_width,&src_height);
                                    GdkPixbuf *pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, false, 8, (int)src_width*pdata->zoom, (int)src_height*pdata->zoom);
                                    poppler_page_render_to_pixbuf(pdata->ppage,0,0,(int)src_width*pdata->zoom,(int)src_height*pdata->zoom,pdata->zoom,0,pixbuf); 
                                    gtk_image_set_from_pixbuf(GTK_IMAGE(context), pixbuf);
                                }
                             }
                             break;
                        case P_PDFZOOM:
                             pdata->zoom=param.ToFloat();
                             if ((pdata->pdoc) && (pdata->ppage)) {
                                double src_width=0;
                                double src_height=0;
                                GTimer      *timer=g_timer_new();
                                poppler_page_get_size(pdata->ppage,&src_width,&src_height);

                                PopplerPage *ppage=poppler_document_get_page(pdata->pdoc,0);

                                GdkPixbuf *pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, false, 8, (int)src_width*pdata->zoom, (int)src_height*pdata->zoom);
                                if (pixbuf) {
                                    poppler_page_render_to_pixbuf(pdata->ppage,0,0,(int)src_width*pdata->zoom,(int)src_height*pdata->zoom,pdata->zoom,0,pixbuf); 
                                }
                                g_timer_stop (timer);
                                g_timer_destroy (timer);
                                gtk_image_set_from_pixbuf(GTK_IMAGE(context), pixbuf);
                             }
                             break;
                        case P_PDFCURRENTPAGELABEL:
                             if (pdata->pdoc) {
                                PopplerPage *ppage=poppler_document_get_page_by_label(pdata->pdoc,str_buffer);
                                pdata->ppage=ppage;
                                if (ppage) {
                                    double src_width=0;
                                    double src_height=0;
                                    poppler_page_get_size(ppage,&src_width,&src_height);
                                    GdkPixbuf *pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, false, 8, (int)src_width*pdata->zoom, (int)src_height*pdata->zoom);
                                    poppler_page_render_to_pixbuf(ppage,0,0,(int)src_width*pdata->zoom,(int)src_height*pdata->zoom,pdata->zoom,0,pixbuf); 
                                    gtk_image_set_from_pixbuf(GTK_IMAGE(context), pixbuf);
                                }
                             }
                             break;
                        case P_PAGEICONVIEW:
                             {
                                int ctrl_id=param.ToInt();
                                int cls_id=0;
                                RemoteInvoker(context,INVOKE_GETCONTROL, ctrl_id, (void **)&pdata->iconview, &cls_id);
                                if (cls_id!=60)
                                    pdata->iconview=NULL;
                                if (pdata->iconview)
                                    pdata->prepared=false;
                             }
                             break;
                        case P_THUMBSIZE:
                             pdata->thumb_size=param.ToInt();
                             break;
                        case P_MARKTEXT:
                             break;
                }
                return E_NOERROR;
            }
            break;
        case INVOKE_GETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char **str_buffer=   va_arg(arg, char **);
                int *buf_len     =   va_arg(arg, int*);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                switch (prop_id) {
                        case P_PDFPAGES:
                            {
                                MyPDFData *pdata=0;
                                PopplerDocument *pdoc=0;
                                RemoteInvoker(context,INVOKE_GETEXTRAPTR,&pdata);

                                if (pdata)
                                    pdoc=pdata->pdoc;

                                if (pdoc)
                                    buffered_text=AnsiString((long)poppler_document_get_n_pages(pdoc));
                                else
                                    buffered_text=(char *)"0";
                                *str_buffer=buffered_text.c_str();
                                *buf_len=buffered_text.Length();
                            }
                            break;
                }
                return E_NOERROR;
            }
            break;
        case INVOKE_SETSTREAM:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                char *target    =   va_arg(arg, char *);
                int len_target  =   va_arg(arg, int);
                char *value     =   va_arg(arg, char *);
                int len_value   =   va_arg(arg, int);
                va_end(arg);
                GError *docerr=0;
                int n_pages=0;

                MyPDFData *pdata=0;
                //PopplerDocument *pdoc=0;
                RemoteInvoker(context,INVOKE_GETEXTRAPTR,&pdata);

                if (!pdata) {
                    pdata=new MyPDFData();
                    RemoteInvoker(context,INVOKE_SETEXTRAPTR,(void*)pdata);
                }
                pdata->Load(value, len_value);

                if (pdata->pdoc)
                    g_object_unref(pdata->pdoc);

                pdata->pdoc=poppler_document_new_from_data(pdata->master_data.c_str(),pdata->master_data.Length()/*value, len_value*/, target, &docerr);
                pdata->PopulateIconView();

                n_pages=poppler_document_get_n_pages(pdata->pdoc);

                if (n_pages>0) {
                    PopplerPage *ppage=poppler_document_get_page(pdata->pdoc,0);
                    pdata->ppage=ppage;

                    if (ppage) {
                        double src_width=0;
                        double src_height=0;
                        poppler_page_get_size(ppage,&src_width,&src_height);
                        GdkPixbuf *pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, false, 8, (int)src_width*pdata->zoom, (int)src_height*pdata->zoom);
                        if (pixbuf) {
                            poppler_page_render_to_pixbuf(ppage,0,0,(int)src_width*pdata->zoom,(int)src_height*pdata->zoom,pdata->zoom,0,pixbuf); 
                            gtk_image_set_from_pixbuf(GTK_IMAGE(context), pixbuf);
                        }
                    }
                }

                return E_NOERROR;
            }
            break;
        case INVOKE_SETSECONDARYSTREAM:
            // not needed
            break;
        case INVOKE_CUSTOMMESSAGE:
            {
            }
            break;
        case INVOKE_CREATE:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void **handler  =   va_arg(arg, void **);
                int *pack_type  =   va_arg(arg, int *);
                va_end(arg);
                *pack_type=(int)2;

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;

                GtkWidget *wrapped=gtk_image_new();
                *handler=wrapped;

                return E_ACCEPTED;
            }
            break;
        case INVOKE_DESTROY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void *handler  =   va_arg(arg, void *);
                va_end(arg);

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;

                return E_ACCEPTED;
            }
            break;
        case INVOKE_INITLIBRARY:
            va_start(arg, INVOKE_TYPE);
            RemoteInvoker     =   va_arg(arg, INVOKER);
            va_end(arg);
            return E_NOERROR;

        case INVOKE_DONELIBRARY:
            return E_NOERROR;

        default:
            return E_NOTIMPLEMENTED;
    }
    return E_NOTIMPLEMENTED;
}


