#ifndef __PRINTER_H
#define __PRINTER_H

#include <gtk/gtk.h>
#include "AnsiList.h"
#include <vector>
#include <list>
#include "ToWriteText.h"


typedef struct {
    int type;
    int page;

    int x1;
    int y1;

    int x2;
    int y2;

    int angle1;
    int angle2;

    bool filled;
} VectorDraw;

gboolean REF_on_draw_page(GtkPrintOperation *operation, GtkPrintContext *print_context,gint page_nr,gpointer userdata);
gboolean REF_on_printoperation_done(GtkPrintOperationResult result, GtkPrintOperation *operation,gpointer userdata);
gboolean REF_on_begin_print(GtkPrintOperation *operation, GtkPrintContext *print_context, gpointer userdata);

class ClientPrinter {
public:
    GtkWidget *Owner;
    GtkUnit Unit;
    GtkPageSetup *PageSetup;
    GtkPrintSettings *Settings;
    GtkPrintOperation *print;

    PangoLayout *m_refLayout;
    std::vector<int> m_PageBreaks;
    //std::vector<int> m_PagePangos;

    AnsiList texts;
    AnsiList new_pages;

    int CurrentPage;
    bool initialized;
    bool print_init;

    int spacing;
    int wrapmode;
    int indent;
    int justify;
    int current_page;

    GdkWindow *backwindow;
    //Cairo::RefPtr<Cairo::Context> page_background;

    //std::list< Glib::RefPtr<Pango::Layout> > PangoArray;

    std::list< PangoLayout * > PangoArray;
    std::list< VectorDraw > vd;

    ClientPrinter(GtkWidget *Owner) {
        print=0;
        this->Owner=Owner;
        this->PageSetup=gtk_page_setup_new();
        this->Settings=gtk_print_settings_new();
        initialized=true;//false;
        print_init=false;
        //PangoArray.reserve(1000);

        spacing=-1;
        wrapmode=-1;
        indent=-1;
        justify=-1;

        current_page=0;
        backwindow=0;
        //page_background=0;
    }
    
    void Draw(int type, int x1=0, int y1=0, int x2=0, int y2=0, int angle1=0, int angle2=0, bool filled=false) {
        VectorDraw _vd;

        if (!current_page)
            current_page=1;

        _vd.type=type;
        _vd.page=current_page;
        _vd.x1=x1;
        _vd.y1=y1;
        _vd.x2=x2;
        _vd.y2=y2;
        _vd.angle1=angle1;
        _vd.angle2=angle2;
        _vd.filled=filled;

        vd.push_back(_vd);
    }

    void SetBackground(GdkWindow *wnd) {

        backwindow=wnd;

        /*if (page_background)
            free(page_background);*/

        //page_background=0;
        /*if (wnd)
            page_background=wnd->create_cairo_context();
        else {
            std::cerr << "ClientPrinter.SetBackground: Window is invalid (is null)\n";
        }*/
    }

    void InitPrint() {
        texts.Clear();
        new_pages.Clear();
        vd.clear();
        current_page=0;
    }

    void BeginPrint() {
        //if (print)
        //    g_object_unref(print);
        print = gtk_print_operation_new();

        g_signal_connect(G_OBJECT(print), "begin-print", G_CALLBACK(REF_on_begin_print), this);
        g_signal_connect(G_OBJECT(print), "draw-page", G_CALLBACK(REF_on_draw_page), this);

        gtk_print_operation_set_default_page_setup(print, PageSetup);
        gtk_print_operation_set_print_settings(print, Settings);

        gtk_print_operation_set_track_print_status(print, TRUE);

        print_init=true;
    }


    void on_printoperation_done(GtkPrintOperationResult result, GtkPrintOperation *operation) {
        //Printing is "done" when the print data is spooled.

        if (result == GTK_PRINT_OPERATION_RESULT_ERROR)
        {
            //Gtk::MessageDialog err_dialog(*(Gtk::Window *)Owner, "Error printing form", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
            //err_dialog.run();
        }
        else if (result == GTK_PRINT_OPERATION_RESULT_APPLY)
        {
            //Update PrintSettings with the ones used in this PrintOperation:
            Settings = gtk_print_operation_get_print_settings(operation);
        }

        if (!gtk_print_operation_is_finished(operation))
        {
            //We will connect to the status-changed signal to track status
            //and update a status bar. In addition, you can, for example,
            //keep a list of active print operations, or provide a progress dialog.
            //operation->signal_status_changed().connect(sigc::bind(sigc::mem_fun(*this, &ExampleWindow::on_printoperation_status_changed), operation));
        }
    }

    void on_draw_page(GtkPrintOperation *operation, GtkPrintContext *print_context,gint page_nr) {
        //Decide which lines we need to print in order to print the specified page:
        int start_page_line = 0;
        int end_page_line = 0;

        if(page_nr == 0) {
            start_page_line = 0;
        } else {
            start_page_line = m_PageBreaks[page_nr - 1];
        }

        int end_page_pango=0;

        if(page_nr < static_cast<int>(m_PageBreaks.size())) {
            end_page_line = m_PageBreaks[page_nr];
        } else {
            if (PangoArray.size()) {
                int sz=PangoArray.size();
                end_page_pango = 0;

                std::list< PangoLayout* >::const_iterator it;
                int i=0;
                for (it=PangoArray.begin(); it!=PangoArray.end(); ++it)
                    end_page_line+=pango_layout_get_line_count((*it));
            } else
            if (page_nr > current_page)
                return;
        }

        //Get a Cairo Context, which is used as a drawing board:
        cairo_t *cairo_ctx = gtk_print_context_get_cairo_context(print_context);

        //We'll use black letters:
        cairo_set_source_rgb(cairo_ctx, 0, 0, 0);
        //cairo_ctx->set_source_rgb(0, 0, 0);
        //Render Pango LayoutLines over the Cairo context:
        PangoLayoutIter *iter;

        //int cnt=PangoArray.size();
        double start_pos = 0;
        int line_index = 0;

        int base_delta=0;
        int delta=0;


        if (backwindow) {
            cairo_surface_t *image;

            cairo_t *cr = gtk_print_context_get_cairo_context(print_context);

            int dwidth, dheight;
            gdk_drawable_get_size(backwindow,&dwidth, &dheight);

            GdkPixbuf *tmp_buf = gdk_pixbuf_get_from_drawable (NULL,  
                                                    backwindow,
                                                    NULL,  
                                                    0, 0, 
                                                    0, 0, 
                                                    dwidth, dheight);  
            double print_width = gtk_print_context_get_width(print_context);
            double print_height = gtk_print_context_get_height(print_context);

            double scale_w = print_width/PANGO_SCALE;//dwidth  ?  print_width/dwidth   : 1;
            double scale_h = print_height/PANGO_SCALE;//dheight ?  print_height/dheight : 1;

            GdkPixbuf *dest=0;
            
            if (scale_w>scale_h)
                dest=gdk_pixbuf_scale_simple(tmp_buf, (int)(dwidth*scale_h), (int)(dheight*scale_h), GDK_INTERP_NEAREST);
            else
                dest=gdk_pixbuf_scale_simple(tmp_buf, (int)(dwidth*scale_w), (int)(dheight*scale_w), GDK_INTERP_NEAREST);


            gdk_cairo_set_source_pixbuf(cr, /*tmp_buf*/dest, 0, 0);

            cairo_paint (cr);
            // ?? - crashes here !
            //cairo_surface_destroy (image);
        }
        std::list< PangoLayout* >::const_iterator it;

        int i=0;
        for (it=PangoArray.begin(); it!=PangoArray.end(); ++it) {
        //for (int i=0;i<cnt;i++) {
            i++;

            m_refLayout=*it;//PangoArray[i];
            iter=pango_layout_get_iter(m_refLayout);

            delta=base_delta;
              
            //std::cerr << " DELTA : " << delta  << "\n";
            do {
                if (line_index >= start_page_line)
                {
                    PangoLayoutLine *layout_line = pango_layout_iter_get_line(iter);
                    //PangoRectangle *logical_rect = pango_layout_iter_get_line_logical_extents(iter);
                    PangoRectangle ink_rect, logical_rect;
                    pango_layout_line_get_extents(layout_line, &ink_rect, &logical_rect);
                    int baseline = pango_layout_iter_get_baseline(iter) + delta;
                    //int baseline = delta ? delta : iter.get_baseline()/* + delta*/;

                    base_delta=baseline;

                    if (line_index == start_page_line) {
	                    start_pos = logical_rect.y / 1024.0;
                    }

                    //std::cerr << " BASE LINE : " << baseline  << "\n";

                    cairo_move_to(cairo_ctx, logical_rect.x / 1024.0, baseline / 1024.0 - start_pos);

                    //std::cerr << " BASE DELTA1 : " << base_delta  << "\n";
                    base_delta+=(int)(logical_rect.height/4)/* - (start_pos * 1024)*/;
                    //std::cerr << " BASE DELTA2 : " << base_delta  << "\n";

                    pango_cairo_show_layout_line(cairo_ctx, layout_line);
                }

                line_index++;
            } while (line_index < end_page_line && pango_layout_iter_next_line(iter));

            if ((line_index >= end_page_line) /*&& (i==end_page_pango)*/) {
                break;
            }
        }

        if (current_page) {
            cairo_set_source_rgb(cairo_ctx, 1.0,1.0,1.0);
            cairo_rectangle(cairo_ctx,0,0,gtk_print_context_get_width(print_context), gtk_print_context_get_height(print_context));
            cairo_fill(cairo_ctx);
            //cairo_ctx->clip();
            cairo_stroke(cairo_ctx);
            cairo_set_line_width(cairo_ctx,1);
        }

        std::list< VectorDraw >::const_iterator it2;
        for (it2=vd.begin(); it2!=vd.end(); ++it2) {
            VectorDraw vd=*it2;

            if (vd.page!=page_nr)
                continue;

            switch (vd.type) {
                case MSG_D_PRIMITIVE_LINE:
                    cairo_move_to(cairo_ctx, vd.x1, vd.y1);
                    cairo_line_to(cairo_ctx, vd.x2, vd.y2);
                    cairo_stroke(cairo_ctx);
                    break;
                case MSG_D_PRIMITIVE_ARC:
                case MSG_D_PRIMITIVE_FILL_ARC:
                    {
                        int max=abs(vd.x2-vd.x1);
                        int max2=abs(vd.y2-vd.y1);
                        max=max2 > max ? max2 : max;

                        cairo_arc(cairo_ctx,vd.x1,vd.y1,max,vd.angle1,vd.angle2);
                        if (vd.filled) {
                            cairo_close_path(cairo_ctx);
                            cairo_fill(cairo_ctx);
                        }
                        cairo_stroke(cairo_ctx);
                    }
                    break;
                case MSG_D_PRIMITIVE_RECT:
                case MSG_D_PRIMITIVE_FILL_RECT:
                    cairo_rectangle(cairo_ctx,vd.x1, vd.y1, vd.x2, vd.y2);
                    if (vd.filled)
                        cairo_fill(cairo_ctx);
                    cairo_stroke(cairo_ctx);
                    break;
                case MSG_D_PRIMITIVE_POINT:
                    cairo_move_to(cairo_ctx, vd.x1, vd.y1);
                    cairo_line_to(cairo_ctx, vd.x1+0.1, vd.y1);
                    cairo_stroke(cairo_ctx);
                    break;
                case MSG_D_PRIMITIVE_COLOR:
                    cairo_set_source_rgb(cairo_ctx, (double)vd.x1/0xFF, (double)vd.y1/0xFF, (double)vd.x2/0xFF);
                    break;
                default:
                    std::cerr << "Unimplemented message in printer canvas : " << vd.type << "\n";
            }
        }
        //cairo_ctx->paint();
    }

    void on_begin_print(GtkPrintOperation *operation, GtkPrintContext *print_context) {

        //Create and set up a Pango layout for PrintData based on the passed PrintContext:
        //We then use this to calculate the number of pages needed,
        //and the lines that are on each page.

        while (PangoArray.size())
            PangoArray.pop_back();

        while (m_PageBreaks.size())
            m_PageBreaks.pop_back();

        /*while (m_PagePangos.size())
            m_PagePangos.pop_back();*/

        int len=texts.Count();

        /*if (len>PangoArray.capacity()) {
            PangoArray.resize(len);
        }*/
        PangoArray.clear();

        const double width = gtk_print_context_get_width(print_context);
        const double height = gtk_print_context_get_height(print_context);

        double page_height = 0;
        int real_line=0;
        for (int i=0;i<len;i++) {

            m_refLayout = gtk_print_context_create_pango_layout(print_context);
            PangoArray.push_back(m_refLayout);


            PangoFontDescription *font_desc=pango_font_description_from_string("sans 10");
            pango_layout_set_font_description(m_refLayout, font_desc);
            //g_object_unref(font_desc);

            pango_layout_set_width(m_refLayout, static_cast<int>(width * PANGO_SCALE));

            //Set and mark up the text to print:


            ToWriteText *txt=(ToWriteText *)texts[i];

            if (txt->spacing!=-1)
                pango_layout_set_spacing(m_refLayout,txt->spacing);
            if (txt->wrapmode!=-1)
                pango_layout_set_wrap(m_refLayout,(PangoWrapMode)txt->wrapmode);
            if (txt->indent!=-1)
                pango_layout_set_indent(m_refLayout,txt->indent);
            if (txt->justify!=-1)
                pango_layout_set_justify(m_refLayout,txt->justify);


            pango_layout_set_markup(m_refLayout,txt->text.c_str(), txt->text.Length());

            //Set the number of pages to print by determining the line numbers
            //where page breaks occur:
            const int line_count = pango_layout_get_line_count(m_refLayout);

            PangoLayoutLine *layout_line;

            for (int line = 0; line < line_count; ++line)
            {
                PangoRectangle ink_rect, logical_rect;

                layout_line = pango_layout_get_line(m_refLayout, line);

                pango_layout_line_get_extents(layout_line, &ink_rect, &logical_rect);

                const double line_height = logical_rect.height / 1024.0;

                /*std::cerr << "LINE INDEX  : " << real_line << "\n";
                std::cerr << "PAGE HEIGHT : " << page_height << "\n";
                std::cerr << "LINE HEIGHT : " << line_height << "\n";
                std::cerr << "HEIGHT      : " << height << "\n\n";*/

                if (page_height + line_height > height) {
                    //std::cerr << "===================================================";
                    m_PageBreaks.push_back(/*line*/real_line);
                    //m_PagePangos.push_back(i);
                    page_height = 0;
                }

                page_height += line_height;

                real_line++;
            }
        }
        int pages=m_PageBreaks.size() + 1;
        if ((current_page) && (pages<current_page))
            pages=current_page;

        gtk_print_operation_set_n_pages(print, pages);
    }

    void Write(char *data) {
        ToWriteText *txt=new ToWriteText();
        txt->spacing=spacing;
        txt->wrapmode=wrapmode;
        txt->indent=indent;
        txt->justify=justify;
        txt->text=data;
        texts.Add(txt, DATA_WRITETEXT);
    }

    void NewPage() {
        current_page++;
        //new_pages.Add(new AnsiString((long)texts.Count()), DATA_STRING);
    }

    int Print(GtkPrintOperationAction print_action, char *exprt) {
        BeginPrint();

        if (print_action==GTK_PRINT_OPERATION_ACTION_EXPORT)
            gtk_print_operation_set_export_filename(print, exprt);

        GError *err=0;
        gtk_print_operation_run(print, print_action, GTK_WINDOW(Owner), &err);

        InitPrint();
        if (err) {
            //See documentation for exact Gtk::PrintError error codes.
            std::cerr << "An error occurred while trying to run a print operation." << std::endl;
            return 0;
        }
        return 1;
    }

    void setup_page(int type) {
        if (!type) {
            PageSetup=gtk_print_run_page_setup_dialog(GTK_WINDOW(Owner), PageSetup, Settings);
        } else {

        }
        initialized=true;
    }
};

gboolean REF_on_draw_page(GtkPrintOperation *operation, GtkPrintContext *print_context,gint page_nr,gpointer userdata) {
    ((ClientPrinter *)userdata)->on_draw_page(operation, print_context, page_nr);
    return FALSE;
}

gboolean REF_on_printoperation_done(GtkPrintOperationResult result, GtkPrintOperation *operation,gpointer userdata) {
    ((ClientPrinter *)userdata)->on_printoperation_done(result, operation);
    return FALSE;
}

gboolean REF_on_begin_print(GtkPrintOperation *operation, GtkPrintContext *print_context, gpointer userdata) {
    ((ClientPrinter *)userdata)->on_begin_print(operation, print_context);
    return FALSE;
}

#endif //__PRINTER_H
